
# python
from urllib import quote
from xml.dom import minidom as minidom

# zope
from zope.interface import implements

from anz.casclient.interfaces import IProxyRetriever
from anz.casclient.utils import retrieveResponseFromServer

class Cas20ProxyRetriever( object ):
    ''' Implementation of a ProxyRetriever that follows the CAS 2.0.
    In general, this class will make a call to the CAS server with
    some specified parameters and receive an XML response to parse.
    
    '''
    
    implements( IProxyRetriever )
    
    CAS_NS = 'http://www.yale.edu/tp/cas'
    
    def __init__( self, casServerUrl ):
        ''' Construct a proxy retriever object.
        
        @param casServerUrl
        the URL to the CAS server
        
        '''
        self.casServerUrl = casServerUrl
    
    def getProxyTicketIdFor( self, pgtId, targetService ):
        ''' See interfaces.IProxyRetriever. '''
        ret = None
        
        url = self._constructUrl( pgtId, targetService )
        response = retrieveResponseFromServer( url )
        dom = minidom.parseString( response )
        elements = dom.getElementsByTagNameNS( self.CAS_NS,
                                               'proxyFailure' )
        if not elements:
            elements = dom.getElementsByTagNameNS( self.CAS_NS,
                                                   'proxyTicket' )
            ret = elements[0].firstChild.data
        
        return ret
    
    def _constructUrl( self, pgtId, targetService ):
        url = []
        url.append( self.casServerUrl )
        if not self.casServerUrl[-1] == '/':
            url.append( '/' )
        
        url.append( 'proxy?' )
        url.append( 'pgt=%s&' % pgtId )
        url.append( 'targetService=%s' % quote(targetService) )
        
        return ''.join( url )




# python
import datetime

# zope
from Persistence import Persistent
from zope.interface import implements
from zope.bforest.periodic import OOBForest
import transaction

from anz.casclient.interfaces import IProxyGrantingTicketStorage

class ProxyGrantingTicketStorage( Persistent ):
    ''' See interfaces.IProxyGrantingTicketStorage. '''
    
    implements( IProxyGrantingTicketStorage )
    
    # pgtIou validate in almost 60 seconds before it is considered expired.
    TIME_OUT = 60
    
    def __init__( self ):
        ''' Creates a proxy granting ticket storage. '''
        self._mapping = OOBForest(
            datetime.timedelta(seconds=int(self.TIME_OUT/10)), count=11 )
    
    def add( self, pgtIou, pgt ):
        ''' See interfaces.IProxyGrantingTicketStorage. '''
        self._mapping[pgtIou] = pgt
        
        # Commit explicity to make sure the pgt has been saved before we tried
        # to get it during service validate. This will be a issue when anz.cas
        # and anz.casclient deploied on the same zope instance.
        #transaction.commit()
    
    def retrieve( self, pgtIou ):
        ''' See interfaces.IProxyGrantingTicketStorage. '''
        return pgtIou in self._mapping.keys() and self._mapping[pgtIou] or None




# zope
from Products.Five import BrowserView
from zope.interface import implements

from anz.casclient.utils import retrieveResponseFromServer
from anz.casclient.proxyauthexample.interfaces import IProxyAuthExampleView

class ProxyAuthExampleView( BrowserView ):
    ''' An example view to show how to use proxy authentication. '''
    
    implements( IProxyAuthExampleView )
    
    def __init__( self, context, request ):
        super(ProxyAuthExampleView, self).__init__( context, request )
        
        # eg. http://xx.xx.xx.xx:8080/backend
        #self.BACK_END_SERVICE_URL = '{URL OF YOUR BACK-END SERVICE}'
        self.BACK_END_SERVICE_URL = 'http://cas.server:5150/backend'
        
        # eg. /plone/acl_users/anz_casclient
        #self.PATH_TO_PROXIER_PLUGIN = \
        #    '{PATH TO YOUR PROXIER ANZ.CASCLIENT PLUGIN}'
        self.PATH_TO_PROXIER_PLUGIN = '/plone/acl_users/anz_casclient'
        
        # eg. /backend/acl_users/anz_casclient
        #self.PATH_TO_BACK_END_PLUGIN = \
        #    '{PATH TO YOUR BACK END ANZ.CASCLIENT PLUGIN}'
        self.PATH_TO_BACK_END_PLUGIN = '/backend/acl_users/anz_casclient'
    
    def getUserInfo( self, pt ):
        ''' Running on back-end service to return id of the user whom the
        ticket authenticates.
        
        @param pt
        proxy ticket.
        
        @return
        string message represent running status.
        
        '''
        plugin = self.context.restrictedTraverse(
            self.PATH_TO_BACK_END_PLUGIN, None )
        if plugin:
            success, assertion = plugin.validateProxyTicket( pt )
            
            if success:
                return 'Hello, %s!' % \
                       ( assertion and assertion.getPrincipal().getId() or None )
            else:
                return 'You are not authorized to get user info.'
        
        return 'Get back-end service anz.casclient plugin fail,check your path setting.'
    
    def getUserInfoFromTargetService( self ):
        ''' Running on proxier service to ask back-end service for user info.
        '''
        request = self.request
        context = self.context
        
        # get cas client plugin instance
        plugin = context.restrictedTraverse(
            self.PATH_TO_PROXIER_PLUGIN, None )
        if plugin:
            assertion = plugin.getAssertion( request.SESSION )
            if not assertion:
                return 'No assertion found.'
            else:
                # get proxying ticket
                pt = assertion.getPrincipal().getProxyTicketFor(
                    self.BACK_END_SERVICE_URL )
                
                # call proxied service
                url = '%s/@@proxyAuthExample/getUserInfo?pt=%s' % \
                    ( self.BACK_END_SERVICE_URL, pt )
                response = retrieveResponseFromServer( url )
                return response
        
        return 'Get proxier service anz.casclient plugin fail, check your path setting.'




# zope
from AccessControl.Permissions import manage_users as ManageUsers

# cmf
from Products.PluggableAuthService.PluggableAuthService import \
     registerMultiPlugin

from anz.casclient.casclient import AnzCASClient, \
     manage_addAnzCASClient, addAnzCASClientForm

# register plugins with pas
try:
    registerMultiPlugin( AnzCASClient.meta_type )
except RuntimeError:
    # make refresh users happy
    pass

def initialize( context ):
    context.registerClass( AnzCASClient,
                           permission=ManageUsers,
                           constructors=( addAnzCASClientForm,
                                          manage_addAnzCASClient
                                          ),
                           icon='www/anz_casclient.png',
                           visibility=None
                           )




# zope
from zope.interface import Interface

class IProxyAuthExampleView( Interface ):
    ''' Interface for the proxy authentication example view. '''
    
    def getUserInfo( ticket ):
        ''' Running on back-end service to return id of the user whom the
        ticket authenticates.
        
        @param ticket
        proxy ticket.
        
        @return
        string message represent running status.
        
        '''
    
    def getUserInfoFromTargetService():
        ''' Running on proxier service to ask back-end service for user info.
        '''




# python
import socket
from urllib2 import urlopen, URLError, HTTPError
from logging import getLogger

from anz.casclient.exceptions import ConnectionException

LOG = getLogger( 'anz.casclient' )

def retrieveResponseFromServer( url ):
    ''' Contacts the CAS Server and retrieve the response.
    '''
    socket.setdefaulttimeout( 5 )
    try:
        response = urlopen( url )
    except HTTPError, e:
        LOG.warning( e )
        raise ConnectionException( 'Error code: %s' % e.code )
    except URLError, e:
        LOG.warning( e )
        raise ConnectionException( 'Fail to connect, %s' % e.reason )
    except Exception, e:
        LOG.warning( e )
        raise
    
    return response.read()




# python
from xml.dom import minidom as minidom
from urllib import quote
from logging import getLogger

# zope
from AccessControl import ClassSecurityInfo
from Acquisition import aq_base
from Globals import InitializeClass
from OFS.Cache import Cacheable
from zope.interface import implements
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
import transaction

# cmf
from Products.PluggableAuthService.plugins.BasePlugin import BasePlugin
from Products.PluggableAuthService.utils import classImplements
from Products.PluggableAuthService.interfaces.plugins import \
        IExtractionPlugin, IChallengePlugin, IAuthenticationPlugin, \
        ICredentialsResetPlugin, ICredentialsUpdatePlugin

from anz.casclient.interfaces import IAnzCASClient
from anz.casclient.assertion import Assertion
from anz.casclient.principal import Principal
from anz.casclient.proxygrantingticketstorage import ProxyGrantingTicketStorage
from anz.casclient.sessionmappingstorage import SessionMappingStorage
from anz.casclient.validationspecification import Cas10TicketValidator, \
     Cas20ServiceTicketValidator, Cas20ProxyTicketValidator
from anz.casclient.exceptions import BaseException
from anz.casclient.utils import retrieveResponseFromServer

LOG = getLogger( 'anz.casclient' )

addAnzCASClientForm = PageTemplateFile(
    'www/add_anzcasclient_form.pt', globals() )

def manage_addAnzCASClient( self, id, title=None, REQUEST=None ):
    ''' Add an instance of anz cas client to PAS. '''
    obj = AnzCASClient( id, title )
    self._setObject( obj.getId(), obj )
    
    if REQUEST is not None:
        REQUEST['RESPONSE'].redirect(
            '%s/manage_workspace'
            '?manage_tabs_message='
            'AnzCentralAuthService+added.'
            % self.absolute_url()
            )

class AnzCASClient( BasePlugin, Cacheable ):
    ''' Anz CAS client, Implement as a PAS plugin. '''
    
    implements( IAnzCASClient )
    
    meta_type = 'Anz CAS Client'
    
    # Session variable use to save assertion
    CAS_ASSERTION = '__cas_assertion'
    
    # The start of the CAS server URL
    casServerUrlPrefix = ''
    
    # An identify of current service.
    # CAS will redirects to here after login.
    # Set this explicitly but not determine it automatically from request makes
    # us get more security assurance.
    # https://wiki.jasig.org/display/CASC/CASFilter
    serviceUrl = ''
    
    # Whether to store the Assertion in session or not.
    # If sessions are not used, proxy granting ticket will be required for
    # each request. Default set to True.
    useSession = True
    
    # If set to True, CAS will ask user for credentials again to authenticate,
    # this may be used for high-security applications. Default set to False.
    renew = False
    
    # If set to True, CAS will not ask the user for credentials. If the
    # user has a pre-existing single sign-on session with CAS, or if a single
    # sign-on session can be established through non-interactive means
    # (i.e. trust authentication), CAS MAY redirect the client to the URL
    # specified by the "service" parameter, appending a valid service ticket.
    # (CAS also MAY interpose an advisory page informing the client that a CAS
    # authentication has taken place.) If the client does not have a single
    # sign-on session with CAS, and a non-interactive authentication cannot be
    # established, CAS MUST redirect the client to the URL specified by the
    # "service" parameter with no "ticket" parameter appended to the URL. If
    # the "service" parameter is not specified and "gateway" is set, the
    # behavior of CAS is undefined. It is RECOMMENDED that in this case, CAS
    # request credentials as if neither parameter was specified.
    # This parameter is not compatible with the "renew" parameter. Behavior
    # is undefined if both are set to True.
    # See details here: http://www.jasig.org/cas/client-integration/gateway
    gateway = False
    
    # Use which CAS protocol to validate ticket
    # one of ['CAS 1.0','CAS 2.0']
    ticketValidationSpecification = 'CAS 1.0'
    ticketValidationSpecification_values = ['CAS 1.0','CAS 2.0']
    
    # The start of the proxy callback url.
    # You should set it point to an instance of this class with protocol 'https'. 
    # The result url will be '{proxyCallbackUrlPrefix}/proxyCallback'.
    # If set, means this service will be used as a proxier to access
    # back-end service on behalf of a particular user.
    proxyCallbackUrlPrefix = ''
    
    # If you provide either the acceptAnyProxy or the allowedProxyChains
    # parameters, a Cas20ProxyTicketValidator will be constructed. Otherwise
    # a Cas20ServiceTicketValidator will be constructed that does not accept
    # proxy tickets.
    
    # Whether any proxy is OK
    acceptAnyProxy = False
    
    # Allowed proxy chains.
    # Each acceptable proxy chain should include a space-separated list of URLs.
    # These URLs are proxier's proxyCallbackUrl.
    allowedProxyChains = []
    
    security = ClassSecurityInfo()
    
    _properties = (
        {
            'id': 'serviceUrl',
            'lable': 'Service URL',
            'type': 'string',
            'mode': 'w'
            },
        {
            'id': 'casServerUrlPrefix',
            'lable': 'CAS Server URL Prefix',
            'type': 'string',
            'mode': 'w'
            },
        {
            'id': 'useSession',
            'lable': 'Use Session',
            'type': 'boolean',
            'mode': 'w'
            },
        {
            'id': 'renew',
            'lable': 'Renew',
            'type': 'boolean',
            'mode': 'w'
            },
        {
            'id': 'gateway',
            'lable': 'Gateway',
            'type': 'boolean',
            'mode': 'w'
            },
        {
            'id': 'ticketValidationSpecification',
            'lable': 'Ticket Validation Specification',
            'select_variable': 'ticketValidationSpecification_values',
            'type': 'selection',
            'mode': 'w'
            },
        {
            'id': 'proxyCallbackUrlPrefix',
            'lable': 'Proxy Callback URL Prefix',
            'type': 'string',
            'mode': 'w'
            },
        {
            'id': 'acceptAnyProxy',
            'lable': 'Accept Any Proxy',
            'type': 'boolean',
            'mode': 'w'
            },
        {
            'id': 'allowedProxyChains',
            'label': 'Allowed Proxy Chains',
            'type': 'lines',
            'mode': 'w'
            },
        )
    
    def __init__( self, id, title ):
        self._id = self.id = id
        self.title = title
        self._pgtStorage = ProxyGrantingTicketStorage()
        self._sessionStorage = SessionMappingStorage()
    
    security.declarePrivate( 'extractCredentials' )
    def extractCredentials( self, request ):
        ''' Extract credentials from session or 'request'. '''
        creds = {}
        
        # Do logout if logout request found
        logoutRequest = request.form.get( 'logoutRequest', '' )
        if logoutRequest:
            self.logoutCallback()
            return creds
        
        sdm = getattr( self, 'session_data_manager', None )
        assert sdm is not None, 'No session data manager found!'
        
        session = sdm.getSessionData( create=0 )
        assertion = self.getAssertion( session )
        if not assertion:
            # Not already authenticated. Is there a ticket in the URL?
            ticket = request.form.get( 'ticket', None )
            if not ticket:
                return None # No CAS authentification
            
            service = self.getService()
            assertion = self.validateServiceTicket( service, ticket )
            
            # Save current user's session to be used by 'Single Sign Out'
            if not session:
                session = sdm.getSessionData( create=1 )
            
            # Get session token as id, it is more reliable
            sessionId = session.getContainerKey()
            self._sessionStorage.addSession( ticket, sessionId )
            
            # Save assertion into session
            if self.useSession:
                # During ticket validation process, the server with this client
                # installed will be callback several times by CAS, this will
                # makes the session data object we get before stale, so here
                # we get it again.
                session = sdm.getSessionData()
                session.set( self.CAS_ASSERTION, assertion )
        
        creds['login'] = assertion.getPrincipal().getId()
        return creds
    
    security.declarePrivate( 'authenticateCredentials' )
    def authenticateCredentials( self, credentials ):
        if credentials['extractor'] != self.getId():
            return None
        
        login = credentials['login']
        return ( login, login )
    
    security.declarePrivate( 'challenge' )
    def challenge( self, request, response, **kw ):
        ''' Challenge the user for credentials. '''
        # Remove current credentials.
        session = request.SESSION
        session[self.CAS_ASSERTION] = None
        
        # Redirect to CAS login URL.
        if self.casServerUrlPrefix:
            url = self.getLoginURL() + '?service=' + self.getService()
            if self.renew:
                url += '&renew=true'
            if self.gateway:
                url += '&gateway=true'
            
            response.redirect( url, lock=1 )
            return 1
        
        # Fall through to the standard unauthorized() call.
        return 0
    
    security.declarePrivate( 'resetCredentials' )
    def resetCredentials( self, request, response ):
        ''' Clears credentials and redirects to CAS logout page. '''
        session = request.SESSION
        session.clear()
        
        if self.casServerUrlPrefix:
            return response.redirect( self.getLogoutURL(), lock=1 )
    
    security.declarePublic( 'proxyCallback' )
    def proxyCallback( self, pgtId=None, pgtIou=None ):
        ''' See interfaces.IAnzCASClient. '''
        ret = 'success'
        if pgtId and pgtIou:
            self._pgtStorage.add( pgtIou, pgtId )
            ret = '<?xml version=\"1.0\"?>'
            ret += '<casClient:proxySuccess xmlns:casClient="http://www.yale.edu/tp/casClient" />'
        
        return ret
    
    security.declarePublic( 'logoutCallback' )
    def logoutCallback( self ):
        ''' See interfaces.IAnzCASClient. '''
        msg = 'No session id found.'
        dom = minidom.parseString( self.REQUEST.form.get('logoutRequest','') )
        SAMLP_NS = 'urn:oasis:names:tc:SAML:2.0:protocol'
        elements = dom.getElementsByTagNameNS( SAMLP_NS,
                                               'SessionIndex' )
        if elements:
            mappingId = elements[0].firstChild.data
            sessionId = self._sessionStorage.getSessionId( mappingId )
            self._sessionStorage.removeByMappingId( mappingId )
            
            sdm = getattr( self, 'session_data_manager', None )
            assert sdm is not None, 'No session data manager found!'
            
            session = sdm.getSessionDataByKey( sessionId )
            if session:
                session.clear()
                
                # We must commit here to make sure the session will be cleared.
                transaction.commit()
            
            msg = 'Logout seccess.'
        
        return msg
    
    security.declarePublic( 'validateProxyTicket' )
    def validateProxyTicket( self, ticket ):
        ''' See interfaces.IAnzCASClient. '''
        validator = Cas20ProxyTicketValidator(
            self.casServerUrlPrefix,
            self._pgtStorage,
            acceptAnyProxy=self.acceptAnyProxy,
            allowedProxyChains=self.allowedProxyChains,
            renew=self.renew )
        
        try:
            assertion = validator.validate( ticket, self.getService() )
        except BaseException, e:
            LOG.warning( e )
            return False, None
        except Exception, e:
            LOG.warning( e )
            return False, None
        else:
            return True, assertion
    
    def getLoginURL( self ):
        ''' See interfaces.IAnzCASClient. '''
        return self.casServerUrlPrefix + '/login'
    
    def getLogoutURL( self ):
        ''' See interfaces.IAnzCASClient. '''
        return self.casServerUrlPrefix + '/logout?url=%s' % self.getService()
    
    def getService( self ):
        ''' See interfaces.IAnzCASClient. '''
        serviceUrl = ''
        
        if self.serviceUrl:
            # use explicitly setted service url
            serviceUrl = self.serviceUrl
        else:
            # extract service URL from REQUEST and remove the ticket parameters.
            request = self.REQUEST
            
            # get query string but strip ticket parameter
            query_string = request.get( 'QUERY_STRING', '' )
            parts = [ p for p in query_string.split('&') if \
                      (p and p.find( 'ticket=' )!=0) ]
            
            url = request.get( 'ACTUAL_URL', request['URL'] )
            
            serviceUrl = '%s?%s' % ( url, '&'.join(parts) )
        
        return quote( serviceUrl )
    
    def getProxyCallbackUrl( self ):
        ''' See interfaces.IAnzCASClient. '''
        return self.proxyCallbackUrlPrefix and \
               '%s/proxyCallback' % self.proxyCallbackUrlPrefix or ''
    
    def getAssertion( self, session ):
        ''' See interfaces.IAnzCASClient. '''
        assertion = None
        
        if self.useSession and session:
            sessionValue = session.get( self.CAS_ASSERTION )
            if isinstance( sessionValue, Assertion ):
                assertion = sessionValue
        
        return assertion
    
    def validateServiceTicket( self, service, ticket ):
        ''' See interfaces.IAnzCASClient. '''
        if self.ticketValidationSpecification == 'CAS 1.0':
            validator = Cas10TicketValidator(
                self.casServerUrlPrefix, self.renew )
        else:
            if self.acceptAnyProxy or self.allowedProxyChains:
                validator = Cas20ProxyTicketValidator(
                    self.casServerUrlPrefix,
                    self._pgtStorage,
                    acceptAnyProxy=self.acceptAnyProxy,
                    allowedProxyChains=self.allowedProxyChains,
                    renew=self.renew )
            else:
                validator = Cas20ServiceTicketValidator(
                    self.casServerUrlPrefix, self._pgtStorage, self.renew )
        
        return validator.validate(
            ticket, service, self.getProxyCallbackUrl() )

classImplements( AnzCASClient, 
                 IExtractionPlugin, 
                 IChallengePlugin, 
                 ICredentialsResetPlugin, 
                 IAuthenticationPlugin )

InitializeClass( AnzCASClient )




# python
import datetime

# zope
from zope.interface import implements

from anz.casclient.interfaces import IAssertion

class Assertion( object ):
    ''' See interfaces.IAssertion. '''
    
    implements( IAssertion )
    
    def __init__( self, principal ):
        ''' Creates a new Assrtion with the supplied Principal.
        
        @param principal
        the Principal object to associate with the Assertion.
        
        '''
        self.principal = principal
    
    def getPrincipal( self ):
        ''' The principal for which this assertion is valid.
        
        @return
        the principal.
        
        '''
        return self.principal




# zope
from zope.interface import implements

from anz.casclient.interfaces import IPrincipal

class Principal( object ):
    ''' See interfaces.IPrincipal. '''
    
    implements( IPrincipal )
    
    def __init__( self, id, pgt=None, proxyRetriever=None ):
        ''' Construct an principal object.
        
        @param id
        the id of principal
        
        @param pgt
        proxy granting ticket id
        
        @param proxyRetriever
        used to retrieve a proxy ticket from a CAS server
        
        '''
        self.id = id
        self.pgt = pgt
        self.proxyRetriever = proxyRetriever
    
    def getId( self ):
        ''' See interfaces.IPrincipal. '''
        return self.id
    
    def getProxyTicketFor( self, service ):
        ''' See interfaces.IPrincipal. '''
        ret = None
        if self.pgt:
            ret = self.proxyRetriever.getProxyTicketIdFor( self.pgt, service )
        
        return ret




class BaseException( Exception ):
    ''' Base class for exceptions in this module. '''
    pass

class TicketValidationException( BaseException ):
    ''' Exception to be thrown when ticket validation fails. '''
    ERROR_CODE = 'INVALID_TICKET'

class InternalException( BaseException ):
    ''' '''
    ERROR_CODE = 'INTERNAL_ERROR'
    
class ConnectionException( BaseException ):
    ''' '''
    ERROR_CODE = 'CONNECTION_ERROR'

class InvalidProxyChainException( BaseException ):
    ''' '''
    ERROR_CODE = 'INVALID_PROXY_CHAIN'



from setuptools import setup, find_packages
import os

version = '1.0.1'

setup(name='anz.casclient',
      version=version,
      description="This is a Zope PAS plugin that authenticates users against a CAS (Central Authentication Service) server.",
      long_description=open("README.txt").read() + "\n" +
                       open(os.path.join("docs", "HISTORY.txt")).read(),
      # Get more strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
        "Framework :: Plone",
        "Programming Language :: Python",
        ],
      keywords='plone cas sso',
      author='jiangdongjin',
      author_email='eastxing@gmail.com',
      url='http://plone.org/products/anz.casclient/',
      license='GPL',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['anz'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
          # -*- Extra requirements: -*-
          'ZODB3>=3.8.3',
          'zope.proxy>=3.4.1',
          'zope.bforest',
      ],
      entry_points="""
      # -*- Entry points: -*-

      [z3c.autoinclude.plugin]
      target = plone
      """,
      setup_requires=["PasteScript"],
      paster_plugins = ["ZopeSkel"],
      )




# zope
from zope.interface import Interface

class IAnzCASClient( Interface ):
    ''' Interface for the CAS client. '''
    
    def getLoginURL():
        ''' Retrieve where to send people for logging in.
        
        '''
    
    def getLogoutURL():
        ''' Retrieve where to send people for logout.
        
        '''
    
    def getService():
        ''' Retrieve identify of current service.
        CAS will redirects to here after login. Set this explicitly but not
        determine it automatically from request makes us get more security
        assurance.
        
        See here:
        https://wiki.jasig.org/display/CASC/CASFilter
        
        '''
    
    def getProxyCallbackUrl():
        ''' Return proxy callback url of current service.
        This is not None when current service act as a proxier to access
        the back-end service on behalf of an authenticated user.
        
        This URL must be HTTPS, and CAS must verify both that the SSL
        certificate is valid and that its name matches that of the service. 
        
        @return
        Proxy callback url, this URL will uniquely and securely identify
        the proxier service that is proxying the client's authentication.
        
        '''
    
    def getAssertion( session ):
        ''' Retrieve assertion from session.
        
        @param session
        session data object to look assertion at it.
        
        @return
        an assertion object contains authentication information or None.
        
        '''
    
    def validateServiceTicket( service, ticket ):
        ''' Validate service ticket.
        
        @param service
        service url
        
        @param ticket
        service ticket
        
        @return
        an assertion object contains authentication information
        if validate success.
        
        '''
    
    def proxyCallback( pgtId=None, pgtIou=None ):
        ''' Callback by CAS server to send pgtId and pgtIou.
        
        If the HTTP GET returns an HTTP status code of 200 (OK), CAS MUST
        respond to the /serviceValidate (or /proxyValidate) request with
        a service response (Section 2.5.2) containing the proxy-granting
        ticket IOU (Section 3.4) within the <cas:proxyGrantingTicket> block.
        If the HTTP GET returns any other status code, excepting HTTP 3xx
        redirects, CAS MUST respond to the /serviceValidate
        (or /proxyValidate) request with a service response that MUST NOT
        contain a <cas:proxyGrantingTicket> block. CAS MAY follow any HTTP
        redirects issued by the pgtUrl. However, the identifying callback
        URL provided upon validation in the <proxy> block MUST be the same
        URL that was initially passed to /serviceValidate (or /proxyValidate)
        as the "pgtUrl" parameter.
        
        @param pgtId
        proxy granting ticket
        
        @param pgtIou
        proxy granting ticket Iou
        
        '''
    
    def validateProxyTicket( ticket ):
        ''' Called by proxied service(back-end service) to validate
        proxy ticket on CAS server.
        
        @param ticket
        proxy ticket wait for validation.
        
        @return
        (True,assertion) if validate success, variable assertion contains
        authentication info.
        
        (False,None) if validate failure.
        
        '''
    
    def logoutCallback( self ):
        ''' Callback by CAS to do 'Single Sign Out'.
        When logout CAS will callback to each of the services that are
        registered with the system and send a POST request with the following:
        <samlp:LogoutRequest ID="[RANDOM ID]" Version="2.0" IssueInstant="[CURRENT DATE/TIME]">
        <saml:NameID>@NOT_USED@</saml:NameID>
        <samlp:SessionIndex>[SESSION IDENTIFIER]</samlp:SessionIndex>
        </samlp:LogoutRequest>
        
        At this moment the session identifier is the same as the CAS Service
        Ticket. Use the session identifier to map back to a session which should
        be terminated.
        
        '''

class IProxyGrantingTicketStorage( Interface ):
    ''' Interface for the storage and retrieval of Proxy Granting Ticket.
    '''
    
    def add( pgtIou, pgt ):
        ''' Add a proxy granting ticket to the backing storage facility.
        
        @param pgtIou
        used as the key
        
        @param pgt
        used as the value
        
        '''
    
    def retrieve( pgtIou ):
        ''' Retrieve a proxy granting ticket based on the proxy granting
        ticket Iou.
        
        Note that implementations are not guaranteed to return the same
        result if retrieve is called twice with the same
        proxyGrantingTicketIou.
        
        @param pgtIou
        used as the key
        
        @return
        the ProxyGrantingTicket Id or None if it can't be found
        
        '''

class IPrincipal( Interface ):
    ''' Represents an authenticated user.
    '''
    
    def getId():
        ''' Returns the unique id for the Principal.
        
        @return
        the unique id for the Principal
        
        '''
    
    def getProxyTicketFor( service ):
        ''' Retrieves a CAS proxy ticket for this specific principal.
        
        @param service
        the service we wish to proxy this user to.
        
        @return
        a String representing the proxy ticket.
        
        '''

class IAssertion( Interface ):
    ''' Represents a response to a validation request.
    '''
    
    def getPrincipal():
        ''' The principal for which this assertion is valid.
        
        @return
        the principal.
        
        '''

class IProxyRetriever( Interface ):
    ''' Interface to abstract the retrieval of a proxy ticket to make the
    implementation a black box to the client.
    
    '''
    
    def getProxyTicketIdFor( pgtId, targetService ):
        ''' Retrieves a proxy ticket for a specific targetService.
        
        @param pgtId
        the Proxy Granting Ticket Id
        
        @param targetService
        the service we want to proxy.
        
        @return
        the ProxyTicket Id if Granted, None otherwise.
        
        '''

class ITicketValidator( Interface ):
    ''' Interface of ticket validator.
    
    '''
    
    def getUrlSuffix():
        ''' Retrieve the endpoint of the validation URL.
        Should be relative (i.e. not start with a "/").
        I.e. validate or serviceValidate.
        
        @return the endpoint of the validation URL.
        
        '''
    
    def validate( ticket, service, proxyCallbackUrl='' ):
        ''' Attempts to validate a ticket for the provided service.
        
        @param ticket
        the ticket to attempt to validate.
        
        @param service
        the service this ticket is valid for.
        
        @param proxyCallbackUrl
        proxy callback url, used in proxy authentication.
        
        @return an assertion from the ticket.
        
        '''
    
    def retrieveResponseFromServer( validationUrl, ticket ):
        ''' Contacts the CAS Server to retrieve the response for the ticket
        validation.
        
        @param validationUrl
        the url to send the validation request to.
        
        @param ticket
        the ticket to validate.
        
        @return
        the response from the CAS server.
        
        '''
    
    def parseResponseFromServer( response ):
        ''' Parses the response from the server into a CAS Assertion.
        
        @param response
        the response from the server, in any format.
        
        @return
        the CAS assertion if one could be parsed from the response.
        
        '''

class ISessionMappingStorage( Interface ):
    ''' Stores the mapping between session id and mapping id(service ticket).
    
    '''
    
    def getSessionId( mappingId ):
        ''' Retrieve session id by mapping id.
        
        @param mappingId
        the mapping id or None.
        
        '''   
    
    def removeByMappingId( mappingId ):
        ''' Remove a session by mapping id.
        
        @param mappingId
        the mapping id.
        
        '''
    
    def removeBySessionId( sessionId ):
        ''' Remove a session by session id.
        
        @param sessionId
        the id of the session.
        
        '''
    
    def addSession( mappingId, sessionId ):
        ''' Add a session to storage.
        
        @param mappingId
        the mapping id.
        
        @param sessionId
        the id of the session.
        
        '''



# python
from datetime import timedelta

# zope
from Persistence import Persistent
from zope.interface import implements
from zope.bforest.periodic import OOBForest

from anz.casclient.interfaces import ISessionMappingStorage

class SessionMappingStorage( Persistent ):
    ''' See interfaces.ISessionMappingStorage. '''
    
    implements( ISessionMappingStorage )
    
    def __init__( self ):
        ''' Creates a proxy granting ticket storage. '''
        # mapping id as key, session id as value
        self._mapping_id_to_session_id = OOBForest(
            timedelta(hours=1), count=3 )
        
        # session id as key, mapping id as value
        self._session_id_to_mapping_id = OOBForest(
            timedelta(hours=1), count=3 )
    
    def getSessionId( self, mappingId ):
        ''' See interfaces.ISessionMappingStorage. '''
        return mappingId in self._mapping_id_to_session_id.keys() and \
               self._mapping_id_to_session_id[mappingId] or None
    
    def removeByMappingId( self, mappingId ):
        ''' See interfaces.ISessionMappingStorage. '''
        if mappingId in self._mapping_id_to_session_id.keys():
            sessionId = self._mapping_id_to_session_id[mappingId]
            
            del self._session_id_to_mapping_id[sessionId]
            del self._mapping_id_to_session_id[mappingId]
    
    def removeBySessionId( self, sessionId ):
        ''' See interfaces.ISessionMappingStorage. '''
        if sessionId in self._session_id_to_mapping_id.keys():
            mappingId = self._session_id_to_mapping_id[sessionId]
            
            del self._session_id_to_mapping_id[sessionId]
            del self._mapping_id_to_session_id[mappingId]
    
    def addSession( self, mappingId, sessionId ):
        ''' See interfaces.ISessionMappingStorage. '''
        self._mapping_id_to_session_id[mappingId] = sessionId
        self._session_id_to_mapping_id[sessionId] = mappingId




# python
from xml.dom import minidom as minidom

# zope
from zope.interface import implements
import transaction

from anz.casclient.interfaces import ITicketValidator
from anz.casclient.principal import Principal
from anz.casclient.assertion import Assertion
from anz.casclient.proxyretriever import Cas20ProxyRetriever
from anz.casclient.utils import retrieveResponseFromServer
from anz.casclient.exceptions import TicketValidationException, \
     InternalException, ConnectionException, InvalidProxyChainException

class TicketValidator( object ):
    ''' Validator that will confirm the validity of a supplied ticket.
    '''
    
    implements( ITicketValidator )
    
    def __init__( self, casServerUrlPrefix, renew=False ):
        ''' Construct a ticket validator object. '''
        self.casServerUrlPrefix = casServerUrlPrefix
        self.renew = renew
    
    def getUrlSuffix( self ):
        ''' See interfaces.ITicketValidator. '''
        raise NotImplementedError
    
    def validate( self, ticket, service, proxyCallbackUrl='' ):
        ''' See interfaces.ITicketValidator. '''
        validationUrl = self._constructValidationUrl(
            ticket, service, proxyCallbackUrl )
        serverResponse = self.retrieveResponseFromServer(
            validationUrl, ticket )
        
        if not serverResponse:
            raise 'The CAS server returned no response.'
        
        return self.parseResponseFromServer( serverResponse )
    
    def retrieveResponseFromServer( self, validationUrl, ticket ):
        ''' See interfaces.ITicketValidator. '''
        return retrieveResponseFromServer( validationUrl )
    
    def parseResponseFromServer( self, response ):
        ''' See interfaces.ITicketValidator. '''
        raise NotImplementedError
    
    def _constructValidationUrl( self, ticket, service, proxyCallbackUrl ):
        ''' Constructs the URL to send the validation request to.
        
        @param ticket
        the ticket to be validated.
        
        @param service
        the service identifier.
        
        @param proxyCallbackUrl
        proxy callback url, used in proxy authentication.
        
        @return
        the fully constructed URL.
     
        '''
        url = []
        url.append( self.casServerUrlPrefix )
        if not self.casServerUrlPrefix[-1] == '/':
            url.append( '/' )
        
        url.append( self.getUrlSuffix() )
        url.append( '?' )
        
        url.append( 'ticket=%s' % ticket )
        url.append( '&service=%s' % service )
        
        if self.renew:
            url.append( '&renew=true' )
        
        if proxyCallbackUrl:
            url.append( '&pgtUrl=%s' % proxyCallbackUrl )
        
        return ''.join( url )

class Cas10TicketValidator( TicketValidator ):
    ''' Implementation of a Ticket Validator that can validate tickets
    conforming to the CAS 1.0 specification.
    
    '''
    def getUrlSuffix( self ):
        ''' See interfaces.ITicketValidator. '''
        return 'validate'
    
    def parseResponseFromServer( self, response ):
        ''' See interfaces.ITicketValidator. '''
        lines = response.split( '\n' )
        first = lines[0]
        if not first=='yes':
            raise 'CAS Server could not validate ticket.'
        
        userId = lines[1]
        return Assertion( Principal( userId ) )

class Cas20ServiceTicketValidator( TicketValidator ):
    ''' Implementation of the Ticket Validator that will validate Service
    Tickets in compliance with the CAS 2.
    
    '''
    CAS_NS = 'http://www.yale.edu/tp/cas'
    
    def __init__( self, casServerUrlPrefix, pgtStorage, renew=False ):
        super(Cas20ServiceTicketValidator, self).__init__( casServerUrlPrefix,
                                                           renew )
        self.pgtStorage = pgtStorage
    
    def getUrlSuffix( self ):
        ''' See interfaces.ITicketValidator. '''
        return 'serviceValidate'
    
    def parseResponseFromServer( self, response ):
        ''' See interfaces.ITicketValidator. '''
        try:
            dom = minidom.parseString( response )
            elements = dom.getElementsByTagNameNS( self.CAS_NS,
                                                   'authenticationFailure' )
            if elements:
                raise TicketValidationException( elements[0].firstChild.data )
            
            elements = dom.getElementsByTagNameNS( self.CAS_NS, 'user' )
            userId = elements and elements[0].firstChild.data or None
            if not userId:
                raise TicketValidationException(
                    'No principal was found in the response.' )
            
            elements = dom.getElementsByTagNameNS( self.CAS_NS,
                                                   'proxyGrantingTicket' )
            pgtIouNode = elements and elements[0] or None
            if pgtIouNode:
                pgtIou = pgtIouNode.firstChild.data
                
                # Explicite call transaction.begin() to sync invalidations
                # for a given transaction, make sure to get the latest pgt
                # storage data. This will be a issue when anz.cas and
                # anz.casclient deploied on the same zope instance.
                transaction.begin()
                
                pgt = self.pgtStorage.retrieve( pgtIou )
                if pgt:
                    principal = Principal(
                        userId, pgt,
                        Cas20ProxyRetriever(self.casServerUrlPrefix) )
                else:
                    raise TicketValidationException(
                        'No pgt found for pgtIou %s.' % pgtIou )
            else:
                principal = Principal( userId )
            
            return Assertion( principal )
        except Exception, e:
            raise InternalException( str(e) )

class Cas20ProxyTicketValidator( Cas20ServiceTicketValidator ):
    ''' Extension Service Ticket validation to validate service tickets and
    proxy tickets.
    
    '''
    def __init__( self, casServerUrlPrefix, pgtStorage, acceptAnyProxy=True,
                  allowedProxyChains=[], renew=False ):
        super(Cas20ProxyTicketValidator, self).__init__( casServerUrlPrefix,
                                                         pgtStorage,
                                                         renew )
        
        self.acceptAnyProxy = acceptAnyProxy
        self.allowedProxyChains = allowedProxyChains
    
    def getUrlSuffix( self ):
        ''' See interfaces.ITicketValidator. '''
        return 'proxyValidate'
    
    def parseResponseFromServer( self, response ):
        ''' See interfaces.ITicketValidator. '''
        assertion = super(Cas20ProxyTicketValidator, self).\
                  parseResponseFromServer( response )
        
        self._validateProxyChain( response )
        
        return assertion
    
    def _validateProxyChain( self, response ):
        dom = minidom.parseString( response )
        elements = dom.getElementsByTagNameNS( self.CAS_NS,
                                               'proxy' )
        proxies = []
        for e in elements:
            proxies.append( e.firstChild.data )
        proxies = ' '.join( proxies )
        
        if self.acceptAnyProxy or not proxies or \
           proxies in self.allowedProxyChains:
            return True
        
        raise InvalidProxyChainException( 'Invalid proxy chain: %s' % proxies )
